/************************************************************************************/
/* Copyright (c) 2003, Outside Echo Ltd. & Hewlett-Packard.                         */
/* All rights reserved.                                                             */  
/*                                                                                  */
/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met:                    */
/*                                                                                  */
/*    * Redistributions of source code must retain the above copyright notice, this */
/*      list of conditions and the following disclaimer.                            */
/*    * Redistributions in binary form must reproduce the above copyright notice,   */
/*      this list of conditions and the following disclaimer in the documentation   */
/*      and/or other materials provided with the distribution.                      */
/*    * Neither the name of Outside Echo Ltd. & Hewlett-Packard                     */
/*      nor the names of its contributors may be used to endorse or promote products*/ 
/*      derived from this software without specific prior written permission.       */
/*                                                                                  */ 
/*                                                                                  */
/*   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS NAMELY     */
/*   OUTSIDE ECHO LTD AND HEWLETT-PACKARD "AS IS" AND ANY                           */
/*   EXPRESS  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED     */
/*   WARRANTIES OF  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/*   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE    */
/*   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL     */
/*   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR     */
/*   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER     */
/*   CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  */
/*   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  */
/*   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  DAMAGE.          */
/*                                                                                  */
/************************************************************************************/
/* Class for Hindi Grapheme to Phoneme Conversion                                   */
/*                                                                                  */
/*                                                                                  */
/* AUTHOR: Partha Pratim Talukdar, Hewlett-Packard Labs (India)                     */
/* Email: (ptpartha@india.hp.com)                                                   */
/* CREATED:  Aug 14, 2003                                                           */
/* MODIFIED: April 27, 2004                                                         */
/* MODIFIED: Feb, 2005 Roger Tucker - added explicit instantiation of Vector        */
/*                                    class for char *                              */
/************************************************************************************/

// #include <vector>

#include "Vector.h"

#include <string>
#include "festival.h"
#include "g2p.h"

using namespace std;

//Roger added:
template class Vector<char *>;

/* Original Constructor */

G2P::G2P(char *ruleFileName, char *mappingFileName, char *lexiconFileName, char *inputFileName, char *outputFileName)
{
   iFP.open(inputFileName);

   if((oFP = fopen(outputFileName, "w")) == NULL)
      Error::PrintAndExit("Output File Could Not be Opened");

   /* Load the character to phoneme mapping into memory. */
   mappingTable.LoadMappingFile(mappingFileName);  

   /* Build the data structure which stores the rule base. */
   ruleBase.CreateRuleBase(ruleFileName);

   /* Set the output file pointer. */
   lexHash.SetOpFilePtr(lexiconFileName, oFP);     
}
      
/* Original Version: This method does the conversion on a word by word basis. 
 * Phonetizes at the sentences level taking one word at a time.
 */
void G2P::Run(char *languageCode, char *ruleModifier)
{
   int  ruleResult;
   int  wordCnt, tmpOffset,compCnt; 
   char Word[MAXWORDLEN];
   std::string tmpStr;
   char *tmpSentence; /* Array used to read in the sentence */
   int  i, l, senCount;
   bool isCompound;

   wordCnt = 0;
   compCnt = 0;
   senCount = 0;

   lexHash.InsertLex();
   
   while(!iFP.eof())
   {
      ++senCount;

      getline(iFP, tmpStr, '\n');

      tmpSentence = (char *) malloc(tmpStr.length()+1);

      strcpy(tmpSentence, tmpStr.c_str());

      if(strlen(tmpSentence) == 0)
      {
	 fprintf(oFP," \n");
		
	 continue;
      }

      tmpSentence[strlen(tmpSentence)] = 0;
      tmpOffset = 0;

      while(tmpOffset < strlen(tmpSentence))
      {
	 /**
	  * Fixed on 24-3-2004. 
	  * Skip if current characters are white spaces.
	  */		  
	 if(tmpSentence[tmpOffset] == ' ' || tmpSentence[tmpOffset] == '\t' || tmpSentence[tmpOffset] == '\r')
	 {
	    ++tmpOffset;
	    continue;
	 }
	       
	 sscanf(tmpSentence + tmpOffset, "%s", Word);
	  
	 isCompound = false;
	 tmpOffset += strlen(Word) + 1;  /* Something more robust is needed here */

	 if(strstr(Word, " ") != NULL)
	    continue;

	 if(strlen(Word) >= MAXWORDLEN) 
	 {
	    /* cout << "Skipping Word :: Longer than allowed length\n"; */
	    continue;	     
	 }
	   
	 Word[strlen(Word)] = 0;

	 /* Look up whether the current word is present in the phonetic 
	  * compound lexicon.
	  */
	 isCompound = lexHash.LookUp(Word);

	 if(isCompound) 
	 {
	    ++compCnt;
	    fprintf(oFP, " #");
	    continue;
	 }
	       
	 i = 0;
	 l = 0;

	 /* Changed on 28-10-03 for enclosing the input words with '#' */

	 l = strlen(Word);
	 inpWord[0] = '#';

	 /* For two # and the ending NULL character. 
	  */
	 if(l < MAXWORDLEN - 3) 
	 {
	    for(i=0; i<l; i++)
	    {
	       inpWord[i+1] = Word[i];
	    }

	    inpWord[i+1] = '#';
	    inpWord[i+2] = 0;
	 }
	 else
	 {
	    cout << "Current word is too long: Skipping this word" << endl;
	    continue;
	 }
	   
	 strcpy(inputWord,inpWord);

	 ++wordCnt;

	 if(strcmp(languageCode, "T") == 0)     /* For any unknown language */
	    TamilSchwaInsertion();

	 ruleResult = ruleBase.Match(inputWord, mappingTable, ruleModifier);

	 if(strcmp(languageCode, "H") == 0)          /* Schwa Deletion: A phenomenon in Hindi */
	    HindiSchwaDeletion();                    
	 else if(strcmp(languageCode, "T") == 0)     /* For Tamil.                            */
	    Print();
	 else if(strcmp(languageCode, "U") == 0)     /* For any unknown language              */	 
	    Print();

	 strcpy(inputWord,"");  /* Clear off the string */
	 strcpy(inpWord,"");    /* Clear off the string */
	 strcpy(Word,"");       /* Clear off the string */
	 phoneStack.clear();    /* Clear off the stack  */

	 if(tmpOffset<strlen(tmpSentence)-1) 
	    fprintf(oFP, " #");
      }

      strcpy(tmpSentence,"");
	
      tmpSentence[0] = 0;

      fprintf(oFP," \n");
   }

   printf("\n\nTotal %d words processed\n", wordCnt);
   printf("\nTotal %d compounds processed\n\n", compCnt);

   fclose(oFP); 
}

void G2P::Print() /* Prints the results to the outputfile */
{
   int i;

   for(i=0; i<ruleBase.phoneticEq.fragmentCount; i++)
   {
      if(strcmp(ruleBase.phoneticEq.phoneStr[i],"Y")==0) 
	 continue;

      fprintf(oFP, "%s ", ruleBase.phoneticEq.phoneStr[i]);
   } 

   fprintf(oFP, "\n");
}

/* This method takes care of Tamil Schwa insertion          */
/* Added by deepa on 30th Oct for the "a" and AB BC problem */

void G2P::TamilSchwaInsertion()
{
   int    i, j, tmpCurIdx, tmpNxtIdx;
   char   tmpCurType, tmpNxtType;

   for(i=0,j=0; i<strlen(inpWord); i++,j++)
   {
      tmpCurIdx =  ReturnInteger(inpWord[i]);
      tmpNxtIdx =  ReturnInteger(inpWord[i+1]);

      tmpCurType = mappingTable.mappingNode[tmpCurIdx].type;
      tmpNxtType = mappingTable.mappingNode[tmpNxtIdx].type;

      inputWord[j] = inpWord[i];
	    	    
      if(tmpCurType == 'C' && tmpNxtType == 'C')
      {
	 inputWord[++j] = 164;  /* ISCII for 'a' */
      }      
   }
	
   /* For the last character which was missed in the for loop */

   inputWord[j] = inpWord[i];    
}

int G2P::ReturnInteger(char ch)
{
   unsigned int idx; 

   idx = (unsigned int) ch;
   idx &= 0x000000ff;	 
	 
   return(idx);
}
 
/* This method handles Hindi schwa deletion. 
 */
void G2P::HindiSchwaDeletion()
{
   int i, tmpIdx;
   bool   halantOrEnd, isWordEnd; /* This flag indicates whether current context is halanth or word end */

   /* Set since the first iteration starts from the end.
    */
   halantOrEnd = true; 
   
   for(i=ruleBase.phoneticEq.fragmentCount-1; i>=0; i--)
   {
      if(i == ruleBase.phoneticEq.fragmentCount-1) 
	 isWordEnd = true;
      else 
	 isWordEnd = false;

      tmpIdx = ruleBase.phoneticEq.characterIdx[i];

      if(mappingTable.mappingNode[tmpIdx].type == 'H') /* Check for halanth */
      {
	 halantOrEnd = true;	    
	 continue;
      }

      if(!halantOrEnd && mappingTable.mappingNode[tmpIdx].type == 'C')
      { 
	 if(i==0)
	 {
	    phoneStack.push_back("a");
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);
	    continue;
	 }
	 /* Added by deepa on oct 16th for "prakAr" problem*/
	 else if(i>=2 && ((strcmp(mappingTable.mappingNode[tmpIdx].className,"RA")==0) && \
			  (mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'H') && \
			  (mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-2]].type == 'C')))
	 {
	    phoneStack.push_back("a");
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-2]);
	    i=i-2;
	    continue;
	 } /* End of code for "prakAr" */
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'V')
	 {
	    if(i>=2 && (mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-2]].type == 'C'))
	    {
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);   /* C */
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]); /* V */
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-2]); /* C */
		    
	       /* Decrement the index, the third decrement is taken care of by the for loop. 
		*/
	       i = i - 2; 
	       continue;
	    }
	    else
	    {
	       /* Added by deepa on Oct 14th diff from dhvani VC.
		*/
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);   /* C */
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]); /* V */
	       i =i - 1;
	       continue;
	    }
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'C')
	 {
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);       /* C */
	    phoneStack.push_back("a");                                   /* a */
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]);     /* C */

	    i = i - 1; /* Decrement the index, the second decrement is taken care of by the for loop */
	    continue;
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'H')
	 {
	    /* Order changed here for better results C A C changed to C C A */
	    phoneStack.push_back("a");                                 /* a */ 
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);     /* C */
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-2]);   /* C */

	    i = i - 2; /* Decrement the index, the third decrement is taken care of by the for loop */
	    continue;
	 }
      }
      else if((!halantOrEnd  || (i==ruleBase.phoneticEq.fragmentCount-1)) && mappingTable.mappingNode[tmpIdx].type == 'V')
      {
	 halantOrEnd=false;

	 if((i==0) || (mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'V'))
	 {
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);		 
	    continue;
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'C')
	 {
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);
   
	    /* Added by deepa on Oct 15 for "pandit" problem */
	    if(mappingTable.mappingNode[tmpIdx].type=='S' && \
	       (mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'C'))
	    {
	       phoneStack.push_back("a");
	    } 

	    /* Added by deepa on Oct 15 for "gaye" problem */
	    if((strcmp(mappingTable.mappingNode[tmpIdx].className,"VS_1")==0) && \
	       (mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'C'))
	    {
	       phoneStack.push_back("a");
	    } 
	
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]); 
 
	    i = i - 1; /* Decrement the index, the second decrement is taken care of by the for loop */
	       
	    continue;
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'H')
	 {
	    /* It is assumed here that the character preceding ahalant will be a consoant
	     * Error checking possible here                                               
	     */
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);   
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-2]);
		  
	    /* Decrement the index, the second decrement is taken care of by the for loop */
	    i = i - 2; 

	    continue;
	 }
      }
      else if(halantOrEnd && (mappingTable.mappingNode[tmpIdx].type == 'C')) /* This case maps to 'Ch' in the Dhvani Algo */
      {
	 halantOrEnd = false; /* Resets the flag */

	 if(i==0)
	 {
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);
	    continue;
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'S') 
	 {
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]);

	    i = i - 1;

	    continue;
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'V')
	 {
	    if(i>=2 && (mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-2]].type == 'C'))
	    {
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);   /* C */
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]); /* V */
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-2]); /* C */
	      
	       /* Decrement the index, the third decrement is taken care of by the for loop */
	       i = i - 2; 

	       continue;
	    }
	    else
	    { 
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);   /* C */
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]); /* V */

	       /* Decrement the index, the third decrement is taken care of by the for loop */
	       i = i - 1; 

	       continue;
	    }
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'C')
	 {
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);   /* C */
	    phoneStack.push_back("a");
	    phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i-1]); /* C */

	    /* Decrement the index, the second decrement is taken care of by the for loop */
	    i = i - 1; 

	    continue;
	 }
	 else if(mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-1]].type == 'H')
	 {
	    /* It is assumed here that the character preceding a halant will be a consoant.
	     * Error checking possible here.
	     */

	    /* Two lookaheads since at i-1 position there is a halant */

	    if(i>=2 && mappingTable.mappingNode[ruleBase.phoneticEq.characterIdx[i-2]].type == 'C') 
	    {
	       /* Added on 28th oct for consonant cluster at the end.
		*/
	       /* Commented again on 2-4-2004 */
	       //if(isWordEnd)  
	       //phoneStack.push_back("*a*");		      
		   
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);     /* C */		    
	       continue;
	    }
	    else
	    {
	       phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);   /* C */
	       continue;
	    }
	 }	    
      }

      /* Added for "hain" problem */

      if(i != 0)
	 phoneStack.push_back(ruleBase.phoneticEq.phoneStr[i]);

      halantOrEnd=false;
   }

   /* Prints the phones in the correct order. 
    */
   for(i=phoneStack.size()-1; i>=0; i--) 
   {
      fprintf(oFP, " %s", phoneStack[i]);
   }
}

/* Just for dummy testing */
LISP Start(LISP inputFile, LISP outputFile)
{
   LISP tmpLispPtr;
   
   cout << "Input: " << inputFile->storage_as.string.data << endl;
   cout << "Output: " << outputFile->storage_as.string.data << endl;

   G2P g2p("src/modules/Hindi/hindi_rules", "src/modules/hpl_hin_g2p/Hindi/hindi_mapping", "src/modules/hpl_hin_g2p/Hindi/new_hindi_lexicon",  inputFile->storage_as.string.data, outputFile->storage_as.string.data);
   
   g2p.Run("H","N");

   printf("HPL Indian Hindi G2P\n");

   return(tmpLispPtr);
}
